#ifdef NULL
#undef NULL
#endif

#include <config.h>
#include <setjmp.h>
#ifdef _WIN32
#define extern __declspec(dllimport) extern
#endif
#include <lisp.h>
extern Lisp_Object Fkeywordp(Lisp_Object object);

#undef extern

#include <dynload.h>
#pragma warning(error : 4013)

#ifndef WINDOWSNT
#define INVALID_HANDLE_VALUE NULL
#include <dlfcn.h>
#endif

#include <ole2.h>
#include <OCIdl.h>
#include <assert.h>

Lisp_Object
  Qwin32com, Qwin32com_interface;

Lisp_Object
  Qsymbol_or_cons_p;

#define CHECK_WIN32COM_OBJ(x) CHECK_TYPE(is_valid_win32com_obj(x), Qwin32com, x)

struct win32com_object
{
  struct vectorlike_header header;
  Lisp_Object type;
  Lisp_Object constants;
  Lisp_Object destructor;
  void (*finalizer)(struct Lisp_Vector *);
  IDispatch *pDispatch;
};

struct win32com_interface_object
{
  struct vectorlike_header header;
  Lisp_Object type;
  Lisp_Object callback;
  void (*finalizer)(struct Lisp_Vector *);
  IUnknown pUnknown[];
};

#define COM_RELEASE(X) (X) ? ((X)->lpVtbl->Release(X)) : 0
#define COM_ADDREF(X) (X) ? ((X)->lpVtbl->AddRef(X)) : 0
#define COM_GET_TYPEATTR(X, Y) ((X)->lpVtbl->GetTypeAttr((X), (Y)))


/* prototypes */
static Lisp_Object com_invoke_0(int nargs, Lisp_Object *args, int wFlags);

static int is_valid_win32com_obj(Lisp_Object obj)
{
    struct win32com_object *o =
	(struct win32com_object *)XPNTR(obj);

    return 
      GENERICP(obj) 
      && EQ(o->type, Qwin32com)
      && o->pDispatch != NULL;
}

static OLECHAR *mb2wc(const char *s)
{
  int size 
    = MultiByteToWideChar(CP_UTF8, 0, s, -1, NULL, 0);
  OLECHAR *pBuf
    = SysAllocStringLen(NULL, size);
  MultiByteToWideChar(CP_UTF8, 0, s, -1, pBuf, size);
  return pBuf;
}

static char *wc2mb(const OLECHAR *p)
{
  int size = WideCharToMultiByte(CP_UTF8, 0, p, -1, NULL, 0, NULL, NULL);
  char *s = xmalloc(size * 2 + 30);
  WideCharToMultiByte(CP_UTF8, 0, p, -1, s, size, NULL, NULL);
  s[size] = 0;
  return s;
}

static Lisp_Object excepinfo2msg(EXCEPINFO *pExInfo)
{
    char error_code[40];
    char *pSource = NULL;
    char *pDescription = NULL;
    Lisp_Object msg[2] = {Qnil, Qnil};
    struct gcpro gcpro1, gcpro2;
    
    if(pExInfo->pfnDeferredFillIn != NULL)
      (*pExInfo->pfnDeferredFillIn)(pExInfo);
    if (pExInfo->bstrSource != NULL)
      pSource = wc2mb(pExInfo->bstrSource);
    if (pExInfo->bstrDescription != NULL)
      pDescription = wc2mb(pExInfo->bstrDescription);
    if(pExInfo->wCode == 0)
      sprintf_s(error_code, sizeof(error_code),
		"COM error scode:%lX in ", pExInfo->scode);
    else 
      sprintf_s(error_code, sizeof(error_code),
		"COM error wcode:%u in ", pExInfo->wCode);
    msg[0] = build_string(error_code);

    GCPRO2(msg[0], msg[1]);

    if(pSource != NULL)
      msg[1] = build_string(pSource);
    else 
      msg[1] = build_string("<Unknown>");
    msg[0] = Fconcat(2, msg);

    msg[1] = build_string(" : ");
    msg[0] = Fconcat(2, msg);

    if(pDescription != NULL)
      msg[1] = build_string(pDescription);
    else
      msg[1] = build_string("<No Description>");
    msg[0] = Fconcat(2, msg);

    xfree(pSource);
    xfree(pDescription);

    SysFreeString(pExInfo->bstrDescription);
    SysFreeString(pExInfo->bstrSource);
    SysFreeString(pExInfo->bstrHelpFile);
    UNGCPRO;
    return msg[0];
}

static void win32com_finalizer(struct Lisp_Vector *v)
{
  struct win32com_object *o =
    (struct win32com_object *)v;

  if (o->pDispatch) {
    if (!NILP(o->destructor)) {
      Lisp_Object *a;
      int n;
      if (CONSP(o->destructor)) {
	int i;
	n = XINT(Flength(o->destructor)) + 1;
	a = alloca(sizeof(Lisp_Object) * n);
	for (i = n; i > 0; i--)
	  a[i] = Felt(o->destructor, i - 1);
      } else {
	n = 2;
	a = alloca(sizeof(Lisp_Object) * n);
	a[1] = o->destructor;
      }
      XSETPSEUDOVECTOR (a[0], o, PVEC_GENERIC);
      com_invoke_0(n, a, DISPATCH_METHOD);
    }
    o->pDispatch->lpVtbl->Release(o->pDispatch);
    o->pDispatch = NULL;
  }
  o->destructor = Qnil;
}

struct win32com_object *
allocate_win32com_object(void)
{
  struct win32com_object *obj =
    ALLOCATE_PSEUDOVECTOR (struct win32com_object,
			   finalizer, PVEC_GENERIC);
  obj->type = Qwin32com;
  obj->constants = Qnil;
  obj->destructor = Qnil;
  obj->finalizer = win32com_finalizer;
  obj->pDispatch = NULL;
  return obj;
}

static void win32com_interface_finalizer(struct Lisp_Vector *v)
{
}

struct win32com_interface_object *
allocate_win32com_interface_object (int size)
{
  struct win32com_interface_object *obj =
    (struct win32com_interface_object *)
    allocate_pseudovector
    ((sizeof(struct win32com_interface_object) + size
      - OFFSETOF (struct Lisp_Vector, contents[0])
      + sizeof(Lisp_Object) - 1)
     / sizeof (Lisp_Object),
     PSEUDOVECSIZE (struct win32com_interface_object, finalizer),
     PVEC_GENERIC);

  obj->type = Qwin32com;
  obj->finalizer = win32com_interface_finalizer;
  
  memset(obj->pUnknown, 0, size);
  return obj;
}

static Lisp_Object load_constant(struct win32com_object *);

DEFUN("win32com-create", Fwin32com_create, Swin32com_create, 1, 2, 0,
      doc:/*Create a win32 com object. ARG should be Com program id or class id.
	    if DESTRUCTOR is symbol or cons, it will be invoked Com methos and args*/)
     (arg, destructor)
     Lisp_Object arg;
     Lisp_Object destructor;
{
  OLECHAR *pBuf;
  CLSID   clsid;
  IDispatch *pDispatch;
  HRESULT hr;
  struct win32com_object *o;
  Lisp_Object obj;
  struct gcpro gcpro1;

  CHECK_STRING(arg);
  CHECK_TYPE(NILP(destructor) || SYMBOLP(destructor) || CONSP(destructor),
    Qsymbol_or_cons_p, destructor);

  pBuf = mb2wc(SDATA(arg));

  /* get CLSID from OLE server name */
  hr = CLSIDFromProgID(pBuf, &clsid);
  if(FAILED(hr)) {
    hr = CLSIDFromString(pBuf, &clsid);
  }
  SysFreeString(pBuf);
  if(FAILED(hr))
    error("%s: COM server not found.", SDATA(arg));
  
  /* get IDispatch interface */
  hr = CoCreateInstance(&clsid, NULL, 
			CLSCTX_INPROC_SERVER | CLSCTX_LOCAL_SERVER,
			&IID_IDispatch, (void**)&pDispatch);

  if(FAILED(hr))
    error("%s: Fail to create win32com object.", SDATA(arg));

  GCPRO1(obj); 
  o = allocate_win32com_object();

  XSETPSEUDOVECTOR (obj, o, PVEC_GENERIC);
  o->constants = load_constant(o);
  o->destructor = destructor;
  o->pDispatch = pDispatch;
  UNGCPRO;
  return obj;
}

DEFUN("win32com-destroy", Fwin32com_destroy, Swin32com_destroy, 1, 1, 0,
      doc:/*Destroy a win32 com OBJ.*/)
     (obj)
     Lisp_Object obj;
{
  struct win32com_object *o =
    (struct win32com_object *)XPNTR(obj);

  CHECK_WIN32COM_OBJ(obj);
  if (o->pDispatch) {
    o->pDispatch->lpVtbl->Release(o->pDispatch);
    o->pDispatch = NULL;
  }
  return Qnil;
}

struct dispsink {
  struct  IDispatchVtbl const *lpVtbl;
  DWORD ref;
  DWORD cookie;
  IConnectionPoint *pConnectionPoint;
  ITypeInfo *pTypeInfo;
};

static STDMETHODIMP
dispsink_QueryInterface(IDispatch *disp,
			REFIID     iid,
			LPVOID*    ppv)
{
  struct dispsink *sink = (struct dispsink *)disp;
  return NOERROR;
}

static STDMETHODIMP_(ULONG)
dispsink_AddRef(IDispatch *disp)
{
  struct dispsink *sink = (struct dispsink *)disp;
  return ++sink->ref;
}

static STDMETHODIMP_(ULONG) 
dispsink_Release(IDispatch *disp)
{
  struct dispsink *sink = (struct dispsink *)disp;
  ++sink->ref;

  if(sink->ref != 0)
    return sink->ref;

  return 0;
}

static STDMETHODIMP
dispsink_GetTypeInfoCount(IDispatch *disp, UINT *count)
{
    *count = 0;
    return NOERROR;
}

static STDMETHODIMP
dispsink_GetTypeInfo(IDispatch *disp,
		     UINT info,
		     LCID lcid,
		     ITypeInfo **pInfo)
{
    *pInfo = NULL;
    return DISP_E_BADINDEX;
}

static STDMETHODIMP
dispsink_GetIDsOfNames(IDispatch *disp,
		       REFIID riid,
		       OLECHAR **szNames,
		       UINT cNames,
		       LCID lcid,
		       DISPID *pDispID)
{
    return DISP_E_UNKNOWNNAME;
}

static STDMETHODIMP
dispsink_Invoke(IDispatch *disp,
		DISPID dispid,
		REFIID riid,
		LCID lcid,
		WORD wFlags,
		DISPPARAMS *pdispparams,
		VARIANT *pvarResult,
		EXCEPINFO *pexcepinfo,
		UINT *puArgErr)
{
    return NOERROR;
}

static const IDispatchVtbl dispsink_vtbl = {
  dispsink_QueryInterface,
  dispsink_AddRef,
  dispsink_Release,
  dispsink_GetTypeInfoCount,
  dispsink_GetTypeInfo,
  dispsink_GetIDsOfNames,
  dispsink_Invoke,
};


DEFUN("win32com-create-interface-dispatch", 
      Fwin32com_create_interface_dispatch, 
      Swin32com_create_interface_dispatch, 1, 1, 0,
      doc:/* Create a win32 sink. FUNCTION is a callback.*/)
     (function)
     Lisp_Object function;
{
  Lisp_Object obj;
  struct win32com_interface_object *o 
    = allocate_win32com_interface_object(sizeof(struct dispsink));
  struct dispsink *sink
    = (struct dispsink*)(o->pUnknown);

  XSETPSEUDOVECTOR (obj, o, PVEC_GENERIC);
  o->callback = function;
  sink->lpVtbl = &dispsink_vtbl;
  return obj;
}      

DEFUN("win32com-create-sink", 
      Fwin32com_create_sink, Swin32com_create_sink, 2, 2, 0,
      doc:/* Create a win32 sink. FUNCTION is a callback and
	     INTERFACE is a win32com-interface object.*/)
     (function, stub)
     Lisp_Object function;
     Lisp_Object stub;
{
  CHECK_TYPE(FUNCTIONP(function), Qinvalid_function, function);
  return Qnil;
}


static void lisp2var(Lisp_Object obj, VARIANT *var);
static void seq2array(SAFEARRAY *ar, Lisp_Object obj,
		      long *ub, long *id, int dim)
{
  for (id[dim] = 0; id[dim] < ub[dim]; id[dim]++) {
    Lisp_Object e = Felt(obj, make_number(id[dim]));
    if (dim > 0)
      seq2array(ar, e, ub, id, dim - 1);
    else {
      VARIANT var;
      lisp2var(e, &var);
      SafeArrayPutElement(ar, id, &var);
    }
  }
}

static void lisp2var(Lisp_Object obj, VARIANT *var)
{
  if (is_valid_win32com_obj(obj)) {
    struct win32com_object *o =
      (struct win32com_object *)XPNTR(obj);
    COM_ADDREF(o->pDispatch);
    V_VT(var) = VT_DISPATCH;
    V_DISPATCH(var) = o->pDispatch;
  } else if (STRINGP(obj)) {
    V_VT(var) = VT_BSTR;
    V_BSTR(var) = mb2wc(SDATA(obj));
  } else if (INTEGERP(obj)) {
    if (XINT(obj) > 0x7fffffff) {
	V_VT(var) = VT_R8;
        V_R8(var) = (double)XINT(obj);
    } else {
      V_VT(var) = VT_I4;
      V_I4(var) = XINT(obj);
    }
  } else if (FLOATP(obj)) {
    V_VT(var) = VT_R8;
    V_R8(var) = XFLOAT_DATA(obj);
  } else if (NILP(obj)) {
    V_VT(var) = VT_BOOL;
    V_BOOL(var) = VARIANT_FALSE;
  } else if (EQ(obj, Qt)) {
    V_VT(var) = VT_BOOL;
    V_BOOL(var) = VARIANT_TRUE;
  } else if (CONSP(obj) 
	     && XINT(Flength(obj)) == 3
	     && INTEGERP(XCAR(obj))
	     && INTEGERP(XCAR(XCDR(obj)))
	     && INTEGERP(XCAR(XCDR(XCDR(obj))))) {
    double t;
    SYSTEMTIME st;
    FILETIME ft;
    LONGLONG ll =
      ((XINT(XCAR(obj)) << 16) + XINT(XCAR(XCDR(obj)))) * 10000000; /* 100 nano sec */
    ll += XINT(XCAR(XCDR(XCDR(obj)))) * 10 + 116444736000000000;
    ft.dwLowDateTime = (DWORD)ll;
    ft.dwHighDateTime = ll >> 32;
    FileTimeToSystemTime(&ft, &st);
    SystemTimeToVariantTime(&st, &t);
    V_VT(var) = VT_DATE;
    V_DATE(var) = t;
  } else if (!NILP(Fsequencep(obj))) {
    HRESULT hr;
    long dim = 1;
    Lisp_Object obj2;
    Lisp_Object zero = make_number(0);
    SAFEARRAYBOUND *ab;
    SAFEARRAY *ar;
    long  *ub, *id;
    int i;
    
    for (obj2 = Felt(obj, zero); 
	 !NILP(Fsequencep(obj2));
	 obj2 = Felt(obj2, zero))
      dim++;
    ab = alloca(sizeof(SAFEARRAYBOUND) * dim);
    ub = alloca(sizeof(long) * dim);
    id = alloca(sizeof(long) * dim);
    if (ab == NULL || ub == NULL || id == NULL)
      error("no memory");
    for (obj2 = obj, i = dim - 1;
	 !NILP(Fsequencep(obj2)); 
	 obj2 = Felt(obj2, zero), i--) {
      ab[i].cElements = XINT(Flength(obj2));
      ab[i].lLbound = 0;
      ub[i] = ab[i].cElements;
      id[i] = 0;
    }
    ar = SafeArrayCreate(VT_VARIANT, dim, ab);
    if (ar == NULL)
      hr = E_OUTOFMEMORY;
    else
      hr = SafeArrayLock(ar);
    if (SUCCEEDED(hr)) {
      seq2array(ar, obj, ub, id, dim-1);
      hr = SafeArrayUnlock(ar);
    }
  } else {
    error("cannot convert to variant");
  }
}

static Lisp_Object var2lisp(VARIANT *pvar);
static Lisp_Object array2vec(SAFEARRAY *ar,
			     const long *lb, const long *ub,
			     int *id, int dim)
{
  Lisp_Object obj = Fmake_vector(make_number(ub[dim] - lb[dim]), Qnil);
  struct gcpro gcpro1;

  GCPRO1(obj);
  for (id[dim] = lb[dim]; id[dim] < ub[dim]; id[dim]++) {
    Lisp_Object o;
    if (dim > 0)
      o = array2vec(ar, lb, ub, id, dim - 1);
    else {
      VARIANT var;
      HRESULT hr
	= SafeArrayPtrOfIndex(ar, id, &V_BYREF(&var));
      if (FAILED(hr))
	error("");
      o = var2lisp(&var);
    }
    ASET(obj, id[dim] - lb[dim], o);
  }
  UNGCPRO;
  return obj;
}

static Lisp_Object var2lisp(VARIANT *pvar)
{
  Lisp_Object obj = Qnil;
  HRESULT hr;
  struct gcpro gcpro1;

  GCPRO1(obj);
  while ( V_VT(pvar) == (VT_BYREF | VT_VARIANT) )
    pvar = V_VARIANTREF(pvar);

  if(V_ISARRAY(pvar)) {
    SAFEARRAY *ar = V_ISBYREF(pvar) ? *V_ARRAYREF(pvar) : V_ARRAY(pvar);
    long i;
    long *pID = NULL, *pLB = NULL, *pUB = NULL;
    VARIANT var;

    Lisp_Object val1 = Qnil, val2 = Qnil;
    int dim = SafeArrayGetDim(ar);

    VariantInit(&var);

    V_VT(&var) = (V_VT(pvar) & ~VT_ARRAY) | VT_BYREF;
    
    pID = alloca(sizeof(long) * dim);
    pLB = alloca(sizeof(long) * dim);
    pUB = alloca(sizeof(long) * dim);
    
    if(!pID || !pLB || !pUB)
      error("momory full");

    for(i = 0; i < dim; ++i) {
      SafeArrayGetLBound(ar, i + 1, &pLB[i]);
      SafeArrayGetUBound(ar, i + 1, &pUB[i]);
    }
    memset(pID, 0, sizeof(sizeof(long) * dim));
    hr = SafeArrayLock(ar);

    if (SUCCEEDED(hr))
      obj = array2vec(ar, pLB, pUB, pID, dim - 1);
    SafeArrayUnlock(ar);

    UNGCPRO;
    return obj;
  }

  switch(V_VT(pvar) & ~VT_BYREF){
  case VT_EMPTY:
    break;
  case VT_NULL:
    break;
  case VT_UI1:
    obj = make_number(V_ISBYREF(pvar) ? *V_UI1REF(pvar) : V_UI1(pvar));
    break;
  case VT_I2:
    obj = make_number(V_ISBYREF(pvar) ? *V_I2REF(pvar) : V_I2(pvar));
    break;
  case VT_I4:
    obj = make_number(V_ISBYREF(pvar) ? *V_I4REF(pvar) : V_I4(pvar));
    break;
  case VT_R4:
    obj = make_float(V_ISBYREF(pvar) ? *V_R4REF(pvar) : V_R4(pvar));
    break;
  case VT_R8:
    obj = make_float(V_ISBYREF(pvar) ? *V_R8REF(pvar) : V_R8(pvar));
    break;
  case VT_BSTR:
    {
      BSTR p = V_ISBYREF(pvar) ? *V_BSTRREF(pvar) : V_BSTR(pvar);
      char *s = wc2mb(p);
      obj = build_string(s);
      xfree(s);
    }
    break;
  case VT_ERROR:
    obj = make_number(V_ISBYREF(pvar) ? *V_ERRORREF(pvar) : V_ERROR(pvar));
    break;
  case VT_BOOL:
    obj = V_ISBYREF(pvar)
      ? *V_BOOLREF(pvar) ? Qt : Qnil
      : V_BOOL(pvar) ? Qt : Qnil;
    break;
  case VT_DISPATCH:
    {
      IDispatch *pDispatch;
      if (V_ISBYREF(pvar))
	pDispatch = *V_DISPATCHREF(pvar);
      else
	pDispatch = V_DISPATCH(pvar);

      if (pDispatch != NULL ) {
	struct win32com_object *o;
	COM_ADDREF(pDispatch);
	o = allocate_win32com_object();
	XSETPSEUDOVECTOR (obj, o, PVEC_GENERIC);
	o->pDispatch = pDispatch;
      }
    }
    break;
  case VT_UNKNOWN:
    {
      /* get IDispatch interface from IUnknown interface */
      IUnknown *punk;
      IDispatch *pDispatch;
      HRESULT hr;
      if (V_ISBYREF(pvar))
	punk = *V_UNKNOWNREF(pvar);
      else
	punk = V_UNKNOWN(pvar);

      if(punk != NULL) {
	hr = punk->lpVtbl->QueryInterface(punk, &IID_IDispatch,
					  (void **)&pDispatch);
	if(SUCCEEDED(hr)) {
	  struct win32com_object *o
	    = allocate_win32com_object();
	  XSETPSEUDOVECTOR (obj, o, PVEC_GENERIC);
	  o->pDispatch = pDispatch;
	}
      }
    }
    break;

  case VT_DATE:
    {
      SYSTEMTIME st;
      if (VariantTimeToSystemTime(V_ISBYREF(pvar) 
				  ? *V_DATEREF(pvar) : V_DATE(pvar),
				  &st) == TRUE) {
	FILETIME ft;
	int hi, lo, ms;
	LONGLONG ll;
	SystemTimeToFileTime(&st, &ft);
	ll = (((LONGLONG)ft.dwHighDateTime) << 32)
	  + ft.dwLowDateTime
	  - 116444736000000000; /* 1970/1/1 00:00 - 1601/1/1 00:00 */ 
	ms = ll / 10 % 1000000;
	lo = (unsigned short)(ll / 10 / 1000000);
	hi = (unsigned short)((ll / 10 / 1000000) >> 16);
	obj = Fcons(ms, Qnil);
	obj = Fcons(lo, obj);
	obj = Fcons(hi, obj);
      }
    }
    break;
  case VT_CY:
  default:
    {
      HRESULT hr;
      VARIANT var;
      VariantInit(&var);
      hr = VariantChangeTypeEx(&var, pvar, 
			       LOCALE_SYSTEM_DEFAULT, 0, VT_BSTR);
      if (SUCCEEDED(hr) && V_VT(&var) == VT_BSTR) {
	BSTR p = V_ISBYREF(&var) ? *V_BSTRREF(&var) : V_BSTR(&var);
	char *s = wc2mb(p);
	obj = build_string(s);
	xfree(s);
      }
      VariantClear(&var);
    }
    break;
  }
  UNGCPRO;
  return obj;
}

static Lisp_Object load_constant(struct win32com_object *o)
{
  HRESULT hr;
  LCID lcid = LOCALE_SYSTEM_DEFAULT;
  ITypeLib *tl;
  unsigned int ndx, count;
  int i;
  Lisp_Object val = Qnil, key = Qnil, tbl = Qnil;
  struct gcpro gcpro1, gcpro2, gcpro3;
  int mx = 0;

  {
    ITypeInfo *ti;
    hr = o->pDispatch->lpVtbl->GetTypeInfo(o->pDispatch, 0, lcid, &ti);
    if (FAILED(hr))
      error("fail to GetTypeInfo");
    
    hr = ti->lpVtbl->GetContainingTypeLib(ti, &tl, &ndx);
    if(FAILED(hr)) {
      COM_RELEASE(ti);
      error("fail to GetContainingTypeLib");
    }
    COM_RELEASE(ti);
    count = tl->lpVtbl->GetTypeInfoCount(tl);
  }

  GCPRO3(val, key, tbl);
  for (i = 0; i < count; i++) {
    ITypeInfo *ti;
    TYPEATTR  *ta;
    int iVar;

    hr = tl->lpVtbl->GetTypeInfo(tl, i, &ti);
    if (FAILED(hr))
      continue;

    hr = COM_GET_TYPEATTR(ti, &ta);
    if(FAILED(hr)) {
      COM_RELEASE(ti);
      continue;
    }

    for(iVar = 0; iVar < ta->cVars; iVar++) {
      BSTR bstr;
      VARDESC   *vd;
      unsigned int len;
      char *p;

      hr = ti->lpVtbl->GetVarDesc(ti, iVar, &vd);
      if(FAILED(hr))
	continue;
      if(vd->varkind == VAR_CONST &&
	 !(vd->wVarFlags & (VARFLAG_FHIDDEN |
			    VARFLAG_FRESTRICTED |
			    VARFLAG_FNONBROWSABLE))) {
	hr = ti->lpVtbl->GetNames(ti, vd->memid, &bstr, 1, &len);
	if(FAILED(hr) || len == 0 || !bstr)
	  continue;

	p = wc2mb(bstr);
	if (strlen(p) > mx)
	  mx = strlen(p);
	key = intern(p);
	xfree(p);
	p = NULL;

	SysFreeString(bstr);
	val = var2lisp(vd->lpvarValue);

#ifdef HASH_ATTR
	if (NILP(tbl)) {
	  Lisp_Object a[] = {
	    QCtest, Qeq,
	  };
	  tbl = Fmake_hash_table(sizeof(a) / sizeof(a[0]), a);
	}
	Fputhash(key, val, tbl);
#else
	tbl = Fcons(Fcons(key, val), tbl);
#endif
      }
      ti->lpVtbl->ReleaseVarDesc(ti, vd);
    }
    ti->lpVtbl->ReleaseTypeAttr(ti, ta);
    COM_RELEASE(ti);
  }
  {
    Lisp_Object a[2] = {
      build_string("XXXXXXXXXXXXXXX %d"),
      make_number(mx)
    };
    Fmessage(sizeof(a) / sizeof(a[0]), a);
  }

  COM_RELEASE(tl);
  UNGCPRO;
  return tbl;
}

static Lisp_Object com_invoke_0(int nargs, Lisp_Object *args, int wFlags)
{
  struct win32com_object *o =
    (struct win32com_object *)XPNTR(args[0]);

  Lisp_Object obj, method, plist;
  HRESULT hr;
  LCID    lcid = LOCALE_SYSTEM_DEFAULT;
  DISPID *DispID;
  DISPPARAMS dp = {0};
  OLECHAR **names = NULL;
  int i, n, j;
  unsigned int argErr = 0;

  EXCEPINFO excepinfo;
  VARIANT result;

  CHECK_WIN32COM_OBJ(args[0]);
  CHECK_SYMBOL(args[1]);

  method = args[1];
  args += 2;
  nargs -= 2;

  for (i = 0; i < nargs; i += dp.cNamedArgs ? 2 : 1) {
    if (dp.cNamedArgs && NILP(Fkeywordp(args[i]))) 
      error("not a keyword");
    if (!NILP(Fkeywordp(args[i]))) {
      int size;
      if (i + 1 < nargs)
	error("wrong number of argment");
      if (dp.cNamedArgs == 0) {
	n = (nargs - i) / 2 
	  + (wFlags & DISPATCH_PROPERTYPUT) ? 1 : 0;
	names = alloca(sizeof(OLECHAR *) * (n + 1));
      }
      names[n - dp.cNamedArgs++] = 
	mb2wc(SDATA(SYMBOL_NAME(args[i])) + 1);
    }
    dp.cArgs++;
  }
  DispID = alloca(sizeof(DISPID) *
		  (dp.cNamedArgs +
		   (wFlags & DISPATCH_PROPERTYPUT) != 0
		   ? 2 : 1));

  if (dp.cNamedArgs > 0
      || (wFlags & DISPATCH_PROPERTYPUT) != 0)
    dp.rgdispidNamedArgs = DispID + 1; 
  if (dp.cNamedArgs == 0)
    names = alloca(sizeof(OLECHAR *));

  {
    Lisp_Object a = SYMBOL_NAME(method);
    Fmessage(1, &a);
  }
  names[0] = mb2wc(SDATA(SYMBOL_NAME(method)));
  
  hr = o->pDispatch->lpVtbl
    ->GetIDsOfNames(o->pDispatch, &IID_NULL,
		    names, dp.cNamedArgs + 1, lcid, DispID);

  for (i = 0; i < dp.cNamedArgs + 1; i++) {
    SysFreeString(names[i]);
    names[i] = NULL;
  }
  if(FAILED(hr))
    error("FAIL to get named argument info");
  dp.rgvarg = alloca(sizeof(VARIANTARG) * dp.cArgs);

  for (i = 0, j = dp.cArgs - 1; i < nargs; i++, j--) {
    if (!NILP(Fkeywordp(args[i])))
      i++;
    VariantInit(&dp.rgvarg[j]);
    lisp2var(args[i], &dp.rgvarg[j]);
  }
  
  if (wFlags & DISPATCH_PROPERTYPUT) {
    if (dp.cArgs == 0)
      return Qnil;

    dp.rgdispidNamedArgs[dp.cNamedArgs++] = DISPID_PROPERTYPUT;
  }


  memset(&excepinfo, 0, sizeof(EXCEPINFO));
  VariantInit(&result);
  hr = o->pDispatch->lpVtbl->Invoke(o->pDispatch, DispID[0], 
				    &IID_NULL, lcid, wFlags, &dp, 
				    &result, &excepinfo, &argErr);
  if (FAILED(hr)) {
    xsignal1(Qerror, excepinfo2msg(&excepinfo));
  }

  obj = var2lisp(&result);
  VariantClear(&result);
  return obj;
}

static Lisp_Object com_invoke(int nargs, Lisp_Object *args, int wFlags)
{
  int first, i;
  Lisp_Object obj = Qnil, tmp = Qnil;
  struct gcpro gcpro1, gcpro2;

  if (!VECTORP(args[1]))
    return com_invoke_0(nargs, args, wFlags);
  
  GCPRO2(obj, tmp);

  for (i = XINT(Flength(args[1])); i > 0; i--) {
    Lisp_Object e = AREF(args[1], i - 1);
    char *name, *p, *ctx;
    CHECK_SYMBOL(e);

    name = xstrdup(SDATA(SYMBOL_NAME(e)));
    while((p = strrchr(name, '.')) != NULL) {
	tmp = intern(p + 1);
	obj = Fcons(tmp, obj);
	*p = 0;
    }
    tmp = intern(name);
    obj = Fcons(tmp, obj);
    xfree(name);
  }

  for (tmp = obj, first = 1;
       XINT(Flength(tmp)) > 1; tmp = XCDR(tmp), first = 0) {
    Lisp_Object a[2] = {args[0], XCAR(tmp)};
    args[0] = com_invoke_0(2, a, DISPATCH_PROPERTYGET);
    if (!first)
      Fwin32com_destroy(a[0]);
  }

  args[1] = XCAR(tmp);
  RETURN_UNGCPRO(com_invoke_0(nargs, args, wFlags));
}

DEFUN("win32com-invoke",
      Fwin32com_invoke, Swin32com_invoke, 2, MANY, 0,
      doc: /* invoke the METHOD of OBJ and return results.*/)
     (nargs, args)
     int nargs;
     Lisp_Object *args;
{
  return com_invoke(nargs, args, DISPATCH_METHOD | DISPATCH_PROPERTYGET);
}

DEFUN("win32com-methods",
      Fwin32com_withevents, Swin32com_withevents, 2, MANY, 0,
      doc: /* .*/)
     (nargs, args)
     int nargs;
     Lisp_Object *args;
{
}

      
DEFUN("win32com-getproperty",
      Fwin32com_getproperty, Swin32com_getproperty, 2, MANY, 0,
      doc: /* get the PROP of OBJ.*/)
     (nargs, args)
     int nargs;
     Lisp_Object *args;
{
  return com_invoke(nargs, args, DISPATCH_PROPERTYGET);
}

DEFUN("win32com-putproperty",
      Fwin32com_putproperty, Swin32com_putproperty, 3, MANY, 0,
      doc: /* put property value.*/)
     (nargs, args)
     int nargs;
     Lisp_Object *args;
{
  return com_invoke(nargs, args, DISPATCH_PROPERTYPUT);
}

DEFUN("win32com-constant", 
      Fwin32com_constant, Swin32com_constant, 2, 3, 0,
      doc:/* Look up ATTR in OBJ and return its associated value.
If ATTR is not found, return DFLT which defaults to nil. */)
     (attr, obj, dflt)
     Lisp_Object attr;
     Lisp_Object obj;
     Lisp_Object dflt;
{
  struct win32com_object *o =
	(struct win32com_object *)XPNTR(obj);
  CHECK_WIN32COM_OBJ(obj);
  CHECK_SYMBOL(attr);
  
#ifdef HASH_ATTR
  return Fgethash(attr, o->constants, dflt);
#else
  {
    Lisp_Object t = Fassq(attr, o->attributes);
    if (NILP(t))
      return dflt;
    return XCAR(XCDR(t));
  }
#endif
}

DEFUN("win32com-constant-names", 
      F32com_constant_names, Swin32com_constant_names, 1, 1, 0,
      doc:/* return attribute names of OBJ. */)
     (obj)
     Lisp_Object obj;
{
  struct win32com_object *o =
	(struct win32com_object *)XPNTR(obj);
  Lisp_Object v = Qnil;

  CHECK_WIN32COM_OBJ(obj);
  if (HASH_TABLE_P(o->constants)) {
    struct Lisp_Hash_Table *h = XHASH_TABLE(o->constants);
    struct gcpro gcpro1;
    int i;

    GCPRO1(v);
    for (i = 0; i < HASH_TABLE_SIZE (h); i++)
      if (!NILP (HASH_HASH (h, i))) 
	v = Fcons(HASH_KEY(h, i), v);
    UNGCPRO;
  }
  return v;
}

void
syms_of_win32com ()
{
#define STR_(s) #s
#define STR(s) STR_(s)
#define DEFSYM_LISP(y, s) do {	\
	y = intern(STR(s));	\
	staticpro(&y);		\
    } while(0)
    DEFSYM_LISP(Qwin32com, win32com);
    DEFSYM_LISP(Qwin32com_interface, win32com-interface);
    DEFSYM_LISP(Qsymbol_or_cons_p, symbol-or-cons-p);

    defsubr(&Swin32com_create);
    defsubr(&Swin32com_destroy);
    defsubr(&Swin32com_constant);
    defsubr(&Swin32com_constant_names);
    defsubr(&Swin32com_invoke);
    defsubr(&Swin32com_putproperty);
    defsubr(&Swin32com_getproperty);

    defsubr(&Swin32com_create_interface_dispatch);

}

static void * initialize_win32com (void *reserved)
{
  if(OleInitialize(NULL) != S_OK){
    return 0;
  }  
  syms_of_win32com ();
  return (void *)(EMACS_INT)1;
}

static void finalize_win32com (void *data)
{
  OleUninitialize();
}

DEFINE_DLLMAIN(initialize_win32com, finalize_win32com)
